# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Terminal Widget."""
# Standard library imports
import json
import os
import sys

# Third-party imports
from qtpy.QtCore import (Qt, QUrl, Slot, QEvent, QTimer, Signal,
                         QObject)
from qtpy.QtGui import QKeySequence
from qtpy.QtWebChannel import QWebChannel
from qtpy.QtWebEngineWidgets import (QWebEnginePage, QWebEngineSettings,
                                     QWebEngineView)
from qtpy.QtWidgets import QMenu, QFrame, QVBoxLayout, QWidget, QApplication
from spyder.config.base import DEV, get_translation
from spyder.config.manager import CONF
from spyder.config.gui import is_dark_interface
from spyder.utils import icon_manager as ima
from spyder.utils.qthelpers import create_action, add_actions

# Local imports
from spyder_terminal.widgets.style.themes import ANSI_COLORS
from spyder_terminal.config import CONF_SECTION


PREFIX = 'spyder_terminal.default.'

# For translations
_ = get_translation('spyder_terminal')


class ChannelHandler(QObject):
    """QWebChannel handler for JS calls."""

    sig_ready = Signal()
    sig_closed = Signal()

    def __init__(self, parent):
        """Handler main constructor."""
        QObject.__init__(self, parent)

    @Slot()
    def ready(self):
        """Invoke signal when terminal prompt is ready."""
        self.sig_ready.emit()

    @Slot()
    def close(self):
        """Invoke signal when terminal process was closed externally."""
        self.sig_closed.emit()


class TerminalWidget(QFrame):
    """Terminal widget."""

    terminal_closed = Signal()
    terminal_ready = Signal()

    def __init__(self, parent, port, path='~', font=None, theme=None,
                 color_scheme=None):
        """Frame main constructor."""
        QWidget.__init__(self, parent)
        url = 'http://127.0.0.1:{0}?path={1}'.format(port, path)
        self.handler = ChannelHandler(self)
        self.handler.sig_ready.connect(lambda: self.terminal_ready.emit())
        self.handler.sig_closed.connect(lambda: self.terminal_closed.emit())
        self.view = TermView(self, parent.CONF,
                             term_url=url, handler=self.handler)
        self.font = font
        self.initial_path = path
        self.theme = theme
        self.color_scheme = color_scheme
        self.parent = parent
        self.shortcuts = self.create_shortcuts()
        layout = QVBoxLayout()
        layout.addWidget(self.view)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setFrameStyle(QFrame.StyledPanel | QFrame.Sunken)
        self.setLayout(layout)

        self.body = self.view.document

        self.handler.sig_ready.connect(self.setup_term)

    def setup_term(self):
        """Setup other terminal options after page has loaded."""
        # This forces to display the black background
        print("\0", end='')
        self.set_font(self.font)
        self.set_dir(self.initial_path)
        self.current_theme = self.set_theme(self.theme, self.color_scheme)
        self.set_scrollbar_style()
        options = self.parent.CONF.options(CONF_SECTION)
        dict_options = {}
        for option in options:
            dict_options[option] = self.parent.get_option(option)
        self.apply_settings(dict_options)

    def create_shortcuts(self):
        """Create the terminal shortcuts."""
        return self.view.create_shortcuts()

    def get_shortcut_data(self):
        """
        Return shortcut data, a list of tuples (shortcut, text, default).

        shortcut (QShortcut or QAction instance)
        text (string): action/shortcut description
        default (string): default key sequence
        """
        return self.view.get_shortcut_data()

    def eval_javascript(self, script):
        """Evaluate Javascript instructions inside view."""
        return self.view.eval_javascript(script)

    def set_scrollbar_style(self):
        """Set terminal scrollbar style."""
        if is_dark_interface():
            self.eval_javascript('addClassStyleToContainer("dark-scroll")')

    def set_dir(self, path):
        """Set terminal initial current working directory."""
        self.eval_javascript('setcwd("{0}")'.format(path))

    def set_font(self, font):
        """Set terminal font via CSS."""
        self.font = font
        self.eval_javascript('fitFont("{0}")'.format(self.font))

    def set_theme(self, theme, color_scheme):
        """Set theme for the terminal."""
        supported_themes = ANSI_COLORS.keys()
        new_theme = {}
        if theme not in supported_themes:
            theme = 'spyder' if color_scheme == 'light' else 'spyder/dark'

        new_theme['background'] = CONF.get('appearance',
                                           '{}/background'.format(theme))
        new_theme['foreground'] = CONF.get('appearance',
                                           '{}/normal'.format(theme))[0]
        new_theme['cursor'] = CONF.get('appearance',
                                       '{}/normal'.format(theme))[0]
        new_theme['cursorAccent'] = CONF.get('appearance',
                                             '{}/ctrlclick'.format(theme))
        new_theme['selection'] = CONF.get('appearance',
                                          '{}/occurrence'.format(theme))
        theme_colors = ANSI_COLORS[theme]
        for color in theme_colors:
            new_theme[color] = theme_colors[color]

        self.eval_javascript('setOption("{}", {})'.format('theme', new_theme))
        self.set_option('fontFamily', CONF.get('appearance', 'font/family'))
        return new_theme

    def get_fonts(self):
        """List terminal CSS fonts."""
        return self.eval_javascript('getFonts()')

    def search_next(self, text, case=False, regex=False, word=False):
        """Search in the terminal for the given regex."""
        search_options = {'regex': regex,
                          'wholeWord': word,
                          'caseSensitive': case}
        return self.eval_javascript(
            'searchNext("{}",{})'.format(text, json.dumps(search_options)))

    def search_previous(self, text, case=False, regex=False, word=False):
        """Search in the terminal for the given regex."""
        search_options = {'regex': regex,
                          'wholeWord': word,
                          'caseSensitive': case}
        return self.eval_javascript(
            'searchPrevious("{}", {})'.format(text,
                                              json.dumps(search_options)))

    def exec_cmd(self, cmd):
        """Execute a command inside the terminal."""
        self.eval_javascript('exec("{0}")'.format(cmd))

    def __alive_loopback(self):
        alive = self.is_alive()
        if not alive:
            self.terminal_closed.emit()
        else:
            QTimer.singleShot(250, self.__alive_loopback)

    def is_alive(self):
        """Check if terminal process is alive."""
        alive = self.eval_javascript('isAlive()')
        return alive

    def set_option(self, option_name, option):
        """Set a configuration option in the terminal."""
        self.eval_javascript('setOption("{}", "{}")'.format(option_name,
                                                            option))

    def apply_settings(self, options):
        """Apply custom settings given an option dictionary."""
        # Bell style option
        if 'sound' in options:
            bell_style = 'sound' if options['sound'] else 'none'
            self.set_option('bellStyle', bell_style)
        # Cursor option
        if 'cursor_type' in options:
            cursor_id = options['cursor_type']
            cursor_choices = {0: "block", 1: "underline", 2: "bar"}
            self.set_option('cursorStyle', cursor_choices[cursor_id])
        if 'color_scheme_name' in options:
            color_scheme = CONF.get('appearance', 'ui_theme')
            theme = CONF.get('appearance', 'selected')
            self.set_theme(theme, color_scheme)
        if 'buffer_limit' in options:
            new_lim = options['buffer_limit']
            self.set_option('scrollback', new_lim)


class TermView(QWebEngineView):
    """XTerm Wrapper."""

    def __init__(self, parent, CONF, term_url='http://127.0.0.1:8070',
                 handler=None):
        """Webview main constructor."""
        super().__init__(parent)
        web_page = QWebEnginePage(self)
        self.setPage(web_page)
        self.source_text = ''
        self.parent = parent
        self.CONF = CONF
        self.shortcuts = self.create_shortcuts()
        self.channel = QWebChannel(self.page())
        self.page().setWebChannel(self.channel)
        self.channel.registerObject('handler', handler)

        self.term_url = QUrl(term_url)
        self.load(self.term_url)

        self.document = self.page()
        try:
            self.document.profile().clearHttpCache()
        except AttributeError:
            pass

        self.initial_y_pos = 0
        self.setFocusPolicy(Qt.ClickFocus)

    def copy(self):
        """Copy unicode text from terminal."""
        self.triggerPageAction(QWebEnginePage.Copy)

    def paste(self):
        """Paste unicode text into terminal."""
        clipboard = QApplication.clipboard()
        text = str(clipboard.text())
        if len(text.splitlines()) > 1:
            eol_chars = os.linesep
            text = eol_chars.join((text + eol_chars).splitlines())
        self.eval_javascript('pasteText({})'.format(repr(text)))

    def clear(self):
        """Clear the terminal."""
        self.eval_javascript('clearTerm()')

    def increase_font(self):
        """Increase terminal font."""
        return self.eval_javascript('increaseFontSize()')

    def decrease_font(self):
        """Decrease terminal font."""
        return self.eval_javascript('decreaseFontSize()')

    def create_shortcuts(self):
        """Create the terminal shortcuts."""
        copy_shortcut = self.CONF.config_shortcut(
            lambda: self.copy(),
            context='terminal',
            name='copy',
            parent=self)
        paste_shortcut = self.CONF.config_shortcut(
            lambda: self.paste(),
            context='terminal',
            name='paste',
            parent=self)
        clear_shortcut = self.CONF.config_shortcut(
            lambda: self.clear(),
            context='terminal',
            name='clear',
            parent=self)
        zoomin_shortcut = self.CONF.config_shortcut(
            lambda: self.increase_font(),
            context='terminal',
            name='zoom in',
            parent=self)
        zoomout_shortcut = self.CONF.config_shortcut(
            lambda: self.decrease_font(),
            context='terminal',
            name='zoom out',
            parent=self)
        return [copy_shortcut, paste_shortcut, clear_shortcut, zoomin_shortcut,
                zoomout_shortcut]

    def get_shortcut_data(self):
        """
        Return shortcut data, a list of tuples (shortcut, text, default).

        shortcut (QShortcut or QAction instance)
        text (string): action/shortcut description
        default (string): default key sequence
        """
        return [sc.data for sc in self.shortcuts]

    def contextMenuEvent(self, event):
        """Override Qt method."""
        copy_action = create_action(
            self, _("Copy text"), icon=ima.icon('editcopy'),
            triggered=self.copy,
            shortcut=self.CONF.get_shortcut(CONF_SECTION, 'copy'))
        paste_action = create_action(
            self, _("Paste text"),
            icon=ima.icon('editpaste'),
            triggered=self.paste,
            shortcut=self.CONF.get_shortcut(CONF_SECTION, 'paste'))
        clear_action = create_action(
            self, _("Clear Terminal"),
            triggered=self.clear,
            shortcut=self.CONF.get_shortcut(CONF_SECTION, 'clear'))
        zoom_in = create_action(
            self, _("Zoom in"), triggered=self.increase_font,
            shortcut=self.CONF.get_shortcut(CONF_SECTION, 'zoom in'))
        zoom_out = create_action(
            self, _("Zoom out"), triggered=self.decrease_font,
            shortcut=self.CONF.get_shortcut(CONF_SECTION, 'zoom out'))
        menu = QMenu(self)
        actions = [self.pageAction(QWebEnginePage.SelectAll),
                   copy_action, paste_action, clear_action, None, zoom_in,
                   zoom_out]
        add_actions(menu, actions)
        menu.popup(event.globalPos())
        event.accept()

    def eval_javascript(self, script):
        """
        Evaluate Javascript instructions inside DOM with the expected prefix.
        """
        script = PREFIX + script
        self.document.runJavaScript("{}".format(script), self.return_js_value)

    def return_js_value(self, value):
        """Return the value of the function evaluated in Javascript."""
        return value

    def wheelEvent(self, event):
        """Catch and process wheel scrolling events via Javascript."""
        delta = event.angleDelta().y()
        self.eval_javascript('scrollTerm({0})'.format(delta))

    def event(self, event):
        """Grab all keyboard input."""
        if event.type() == QEvent.ShortcutOverride:
            self.keyPressEvent(event)
            return True
        return True

    def keyPressEvent(self, event):
        """Qt override method."""
        key = event.key()
        modifiers = event.modifiers()

        if modifiers & Qt.ShiftModifier:
            key += Qt.SHIFT
        if modifiers & Qt.ControlModifier:
            key += Qt.CTRL
        if modifiers & Qt.AltModifier:
            key += Qt.ALT
        if modifiers & Qt.MetaModifier:
            key += Qt.META

        sequence = QKeySequence(key).toString(QKeySequence.PortableText)
        if event == QKeySequence.Paste:
            self.paste()
        elif sequence == self.CONF.get_shortcut(CONF_SECTION, 'copy'):
            self.copy()
        elif sequence == self.CONF.get_shortcut(CONF_SECTION, 'paste'):
            self.paste()
        elif sequence == self.CONF.get_shortcut(CONF_SECTION, 'clear'):
            self.clear()
        elif sequence == self.CONF.get_shortcut(
                CONF_SECTION, 'zoom in'):
            self.increase_font()
        elif sequence == self.CONF.get_shortcut(
                CONF_SECTION, 'zoom out'):
            self.decrease_font()
        else:
            super().keyPressEvent(event)


def test():
    """Plugin visual test."""
    from spyder.utils.qthelpers import qapplication
    app = qapplication(test_time=8)
    term = TerminalWidget(None)
    # term.resize(900, 700)
    term.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Spyder terminal configuration page."""

# Standard library imports
from distutils.version import LooseVersion
import os
import platform
import sys

# Third party imports
from qtpy.QtWidgets import (QVBoxLayout, QGroupBox, QGridLayout, QButtonGroup,
                            QRadioButton, QWidget)
from spyder.api.preferences import PluginConfigPage
from spyder.config.base import get_translation
from spyder.utils.programs import find_program

# Constants
WINDOWS = os.name == 'nt'
UNIX_SHELLS = ['bash', 'sh', 'ksh', 'zsh', 'csh', 'pwsh', 'rbash', 'dash',
               'screen', 'tmux', 'tcsh', 'fish']
WINDOWS_SHELLS = ['cmd', 'powershell']

# For translations
_ = get_translation('spyder_terminal')


class TerminalConfigPage(PluginConfigPage):
    def setup_page(self):
        """Create configuration page."""
        options_layout = QGridLayout()
        # Custom shell options
        shell_group = QGroupBox(_("Terminal shell"))
        shell_layout = QVBoxLayout()
        if WINDOWS:
            self.shells = WINDOWS_SHELLS
        else:
            self.shells = UNIX_SHELLS

        valid_shells = []
        for shell in self.shells:
            if find_program(shell) is not None:
                valid_shells.append(shell)
        valid_shells = zip(valid_shells, valid_shells)
        if WINDOWS:
            default_option = 'cmd'
        elif sys.platform.startswith('linux'):
            default_option = 'bash'
        else:
            mac_ver = LooseVersion(platform.mac_ver()[0])
            if mac_ver >= LooseVersion('10.15.0'):
                # Catalina changed the default shell to zsh
                default_option = 'zsh'
            else:
                default_option = 'bash'
        shell_combo = self.create_combobox(_("Select the shell interpreter:"),
                                           valid_shells, 'shell', restart=True,
                                           default=default_option)
        shell_combo.combobox.setMinimumContentsLength(15)
        shell_layout.addWidget(shell_combo)
        shell_group.setLayout(shell_layout)
        shell_layout.addStretch(1)
        options_layout.addWidget(shell_group)

        # Display preferences
        display_group = QGroupBox(_("Terminal display preferences"))
        display_layout = QVBoxLayout()
        # Custom buffer limit
        self.buffer_sb = self.create_spinbox(_("Buffer limit: "), "",
                                             'buffer_limit', min_=100,
                                             default=1000,
                                             max_=1000000, step=1)
        display_layout.addWidget(self.buffer_sb)
        display_group.setLayout(display_layout)
        display_layout.addStretch(1)
        options_layout.addWidget(display_group)

        # Style preferences
        terminal_group = QGroupBox(_("Terminal style preferences"))
        # Custom bar option
        cursor_choices = [(_("Block"), 0), (_("Underline"), 1), (_("Bar"), 2)]
        self.cursor_combo = self.create_combobox(_("Type of cursor:"),
                                                 cursor_choices,
                                                 'cursor_type')
        self.cursor_combo.combobox.setMinimumContentsLength(15)
        options_layout.addWidget(self.cursor_combo)

        # Custom sound option
        self.sound_cb = self.create_checkbox(
            _("Enable bell sound"), 'sound',
            tip=_("Enable bell sound on terminal"))
        options_layout.addWidget(self.sound_cb)

        terminal_group.setLayout(options_layout)

        layout = QVBoxLayout()
        layout.addWidget(shell_group)
        layout.addWidget(display_group)
        layout.addWidget(terminal_group)
        layout.addStretch(1)

        self.setLayout(layout)



# -*- coding: iso-8859-15 -*-

"""Websocket handling class."""

import logging
import tornado.escape
import tornado.websocket

LOGGER = logging.getLogger(__name__)


class MainSocket(tornado.websocket.WebSocketHandler):
    """Handles long polling communication between xterm.js and server."""

    def initialize(self, close_future=None):
        """Base class initialization."""
        self.close_future = close_future

    def open(self, pid):
        """Open a Websocket associated to a console."""
        LOGGER.info("WebSocket opened: {0}".format(pid))
        self.pid = pid
        self.application.term_manager.start_term(pid, self)
        LOGGER.info("TTY On!")

    def on_close(self):
        """Close console communication."""
        LOGGER.info('TTY Off!')
        LOGGER.info("WebSocket closed: {0}".format(self.pid))
        self.application.term_manager.client_disconnected(self.pid, self)
        if self.close_future is not None:
            self.close_future.set_result(("Done!"))

    def on_message(self, message):
        """Execute a command on console."""
        self.application.term_manager.execute(self.pid, message)

    def on_pty_read(self, text):
        """Read data from pty; send to frontend."""
        self.write_message(text)

    def on_pty_died(self):
        """Close websocket if terminal was closed externally."""
        self.close()



# -*- coding: iso-8859-15 -*-

"""
routes.

======

This module establishes and defines the Web Handlers and Websockets
that are associated with a specific URL routing name. New routing
associations must be defined here.

Notes
-----
For more information regarding routing URL and valid regular expressions
visit: http://www.tornadoweb.org/en/stable/guide/structure.html
"""

import spyder_terminal.server.web as web
import spyder_terminal.server.rest as rest
import spyder_terminal.server.websockets as websockets

# Define new rest associations
REST = [
    (r"/api/terminals", rest.term_rest.MainHandler),
    (r"/api/terminals/(.*)/size", rest.term_rest.ResizeHandler)
]

# Define new websocket routes
WS = [
    (r"/terminals/(.*)", websockets.term_ws.MainSocket)
]

# Define new web rendering route associations
WEB = [
    (r'/', web.main_handler.MainHandler)
]

ROUTES = REST + WS + WEB


def gen_routes(close_future):
    """Return a list of HTML redirection routes."""
    if close_future is not None:
        ws = []
        for route in WS:
            ws.append((route[0], route[1],
                       dict(close_future=close_future)))
        return REST + ws + WEB
    return ROUTES
